#!/bin/bash


function openshift_info {
    echo "+====================================================="
    echo "| Container   : $HOSTNAME"
    echo "| kubernetes  : $KUBERNETES_SERVICE_HOST"
    echo "| home        : $HOME"
    echo "+====================================================="
}

function adei_info {
    echo "ADEI Configuration"
    echo "------------------"
    env | grep 'MYSQL_SERVER'
    env | grep 'ADEI_' | grep -v "_SERVICE_" | grep -v "_PORT" | grep -v "PASSWORD"
}

function adei_debug {
    touch /adei/cfg/.debug
}

function adei_go {
    rm -f /adei/cfg/.debug
}

function adminer_start {
    path=$(dirname $0)
    path=$(cd $path; pwd)
    name=$(basename $0)
    script="${path}/${name}"

    if [ -n "$HOME" ]; then
	[ -d "$HOME" ] || mkdir -p $HOME
	[ -f "$HOME"/.bashrc ] || echo "source $script source" > "$HOME"/.bashrc

        if [ -n "$BZR_EMAIL" -a ! -f "$HOME"/.bazaar/bazaar.conf ]; then
            bzr whoami "$BZR_EMAIL"
            bzr config --scope=bazaar ssl.cert_reqs=none
        fi

        if [ -n "$ADEI_REPOSITORY" -a -n "$ADEI_REPOSITORY_USER" -a ! -f "$HOME"/.bazaar/authentication.conf ]; then
	    host=$(echo $ADEI_REPOSITORY | grep -oP '//\K[^/]*')
	    if [ -n "$ADEI_REPOSITORY_PASSWORD" ]; then
		passline="password=$ADEI_REPOSITORY_PASSWORD"
	    else
		passline="#password="
	    fi

	    cat << EOF > "$HOME"/.bazaar/authentication.conf
[adei]
host=$host
user=$ADEI_REPOSITORY_USER
$passline
EOF
	fi

	openshift_info
	echo
	adei_info
	echo
    fi

    while true ; do
      sleep 3600
    done
}

[ -z "$1" ] && adminer_start
